//Copyright © 2015, Oracle and/or its affiliates. All rights reserved.

package util;


import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.Objects;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.ThreadFactory;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;



/**
 * Utilities used by OEPE code.
 *
 * @author Marcelo Paternostro (marcelo.paternostro@oracle.com)
 */
public final class OEPEUtil {
	private OEPEUtil() {
		// Not instantiable
	}

	public static enum PersistenceAction {
		CREATE,
		SAVE,
		DELETE
	}
/*
	public static final Predicate<Resource> DELETE_RESOURCE_WITHOUT_CONTENT_PREDICATE = new Predicate<Resource>() {
		@Override
		public boolean test(Resource t) {
			return t != null && !t.getContents().isEmpty();
		}
	};
*/
	/**
	 * @param array
	 * @return <code>true</code> if the array is <code>null</code>, has 0 length, or
	 *         if all elements are <code>null</code>
	 */
	public static boolean isEmpty(Object[] array) {
		if (array != null && array.length > 0) {
			for (int i = 0; i < array.length; i++) {
				if (array[i] != null) {
					return false;
				}
			}
		}
		return true;
	}



	/**
	 * @param type
	 * @param elements
	 * @return <code>null</code> or the first element that is an instance of the specified type
	 */
	public static <T> T findFirstElement(Class<T> type, Iterable<?> elements) {
		if (type != null && elements != null) {
			for (Object element : elements) {
				if (type.isInstance(element)) {
					return type.cast(element);
				}
			}
		}
		return null;
	}

	/**
	 * @param type
	 * @param elements
	 * @return a not-<code>null</code> list with the elements that are instances of the specified
	 *         type
	 */
	public static <T> List<T> filter(Class<T> type, Iterable<?> elements) {
		if (type != null && elements != null) {
			List<T> list = new LinkedList<>();

			for (Object element : elements) {
				if (type.isInstance(element)) {
					list.add(type.cast(element));
				}
			}

			if (!list.isEmpty()) {
				return list;
			}
		}
		return new ArrayList<>(0);
	}

	public static ExecutorService newFixedCachedThreadPool(int nThreads) {
		return newFixedCachedThreadPool(nThreads, Executors.defaultThreadFactory());
	}

	public static ExecutorService newFixedCachedThreadPool(int nThreads, ThreadFactory threadFactory) {
		return new ThreadPoolExecutor(
			nThreads/2,
			nThreads,
			60, TimeUnit.SECONDS,
			new ArrayBlockingQueue<Runnable>(nThreads),
			Objects.requireNonNull(threadFactory),
			new ThreadPoolExecutor.CallerRunsPolicy());
	}


	/**
	 * @param iterable
	 * @return a not-<code>null</code>, new list with the elements of the specified iterable
	 */
	public static <T> List<T> toList(Iterable<? extends T> iterable) {
		if (iterable != null) {
			List<T> list = new LinkedList<>();
			for (T element : iterable) {
				list.add(element);
			}

			if (!list.isEmpty()) {
				return new ArrayList<>(list);
			}
		}
		return new ArrayList<>(0);
	}

	public static int indexOf(int[] array, int element) {
		if (array != null && array.length > 0) {
			for (int i = 0; i < array.length; i++) {
				if (element == array[i]) {
					return i;
				}
			}
 		}
		return -1;
	}

	public static int indexOf(Object[] array, Object element) {
		if (array != null && array.length > 0) {
			for (int i = 0; i < array.length; i++) {
				if (Objects.equals(element, array[i])) {
					return i;
				}
			}
 		}
		return -1;
	}

		/**
	 * Read all the bytes from an input stream. The {@code initialSize}
	 * parameter indicates the initial size of the byte[] to allocate.
	 *
	 * <p>This method does not close the input stream.</p>
	 *
	 * @param source
	 * @param initialSize
	 * @return a byte array
	 * @throws IOException
	 */
	/*public static byte[] read(InputStream source, int initialSize) throws IOException {
		int capacity = initialSize;
		byte[] buf = new byte[capacity];
		int nread = 0;
		int rem = buf.length;
		int n;

		// read to EOF which may read more or less than initialSize (eg: file
		// is truncated while we are reading)
		while ((n = source.read(buf, nread, rem)) > 0) {
			nread += n;
			rem -= n;
			assert rem >= 0;
			if (rem == 0) {
				// need larger buffer
				int newCapacity = capacity << 1;
				if (newCapacity < 0) {
					if (capacity == Integer.MAX_VALUE)
					 {
						throw new OutOfMemoryError("Required array size too large"); //$NON-NLS-1$
					}
					newCapacity = Integer.MAX_VALUE;
				}
				rem = newCapacity - capacity;
				buf = Arrays.copyOf(buf, newCapacity);
				capacity = newCapacity;
			}
		}
		return (capacity == nread) ? buf : Arrays.copyOf(buf, nread);
	}
*/

	/**
	 * @param resources
	 * @param deletePredicate a predicate to determine whether resource should be deleted or not
	 * @param monitor
	 * @return not-<code>null</code>, unmodifiable map the schemes of the URIs of the resources as keys and a map of the persistence action to be performed on
	 *         each resource.
	 * @throws InterruptedException
	 */
/*	public static Map<String, Map<PersistenceAction, List<Resource>>> analyzeResources(Collection<? extends Resource> resources, Predicate<Resource> deletePredicate, IProgressMonitor monitor) throws InterruptedException {
		return analyzeResources(resources, deletePredicate, false, monitor);
	}*/

	/**
	 * @param resources
	 * @param deletePredicate a predicate to determine whether resource should be deleted or not
	 * @param platformResourceAsScheme true if 'platform:/resource' (as in 'platform:/resource/foo') should be treated as a scheme
	 * @param monitor
	 * @return not-<code>null</code> map the schemes of the URIs of the resources as keys and a not-empty map of the
	 *         persistence action to be performed on each resource
	 * @throws InterruptedException
	 */
/*	public static Map<String, Map<PersistenceAction, List<Resource>>> analyzeResources(Collection<? extends Resource> resources, Predicate<Resource> deletePredicate, boolean platformResourceAsScheme, IProgressMonitor monitor) throws InterruptedException {
		if (resources != null && !resources.isEmpty()) {
			ProgressMonitorUtil.beginTask(monitor, resources.size());

			Map<String, Map<PersistenceAction, List<Resource>>> schemeMap = new HashMap<>(resources.size());
			for (Resource resource : resources) {
				if (resource != null && resource.getURI() != null) {
					PersistenceAction action;

					URIConverter uriConverter = getURIConverter(resource);
					URI uri = uriConverter.normalize(resource.getURI());
					if (!uri.isRelative()) {
						if (uriConverter.exists(uri, null)) {
							action =
								deletePredicate != null && deletePredicate.test(resource) ? PersistenceAction.DELETE :
								PersistenceAction.SAVE;
						} else {
							action =
								resource.getContents().isEmpty() ? null :
								PersistenceAction.CREATE;
						}

						if (action != null) {
							String scheme =
								platformResourceAsScheme && uri.isPlatformResource() ? "platform:/resource" : //$NON-NLS-1$
								uri.scheme();

							Map<PersistenceAction, List<Resource>> actionMap = schemeMap.get(scheme);
							if (actionMap == null) {
								actionMap = new EnumMap<>(PersistenceAction.class);
								schemeMap.put(scheme, actionMap);
							}

							List<Resource> list = actionMap.get(action);
							if (list == null) {
								list = new LinkedList<>();
								actionMap.put(action, list);
							}

							list.add(resource);
						}
					}
				}

				ProgressMonitorUtil.checkIfCanceled(monitor);
				ProgressMonitorUtil.worked(monitor,1);
			}

			ProgressMonitorUtil.done(monitor);
			return schemeMap;
		}
		return new HashMap<>(0);
	}
*/
/*	public static void save(Collection<? extends Map<PersistenceAction, ? extends Collection<? extends Resource>>> actionMaps, IProgressMonitor monitor) throws Exception {
		if (actionMaps != null && !actionMaps.isEmpty()) {
			ProgressMonitorUtil.beginTask(monitor, actionMaps.size());

			for (Map<PersistenceAction, ? extends Collection<? extends Resource>> actionMap : actionMaps) {
				IProgressMonitor subMonitor = ProgressMonitorUtil.submon(monitor, 1);
				ProgressMonitorUtil.beginTask(subMonitor, actionMap.size());

				for (Map.Entry<PersistenceAction, ? extends Collection<? extends Resource>> entry : actionMap.entrySet()) {
					save(entry.getKey(), entry.getValue(), ProgressMonitorUtil.submon(subMonitor, 1));
				}

				ProgressMonitorUtil.done(subMonitor);
				ProgressMonitorUtil.checkIfCanceled(monitor);
			}

			ProgressMonitorUtil.done(monitor);
		}
	}

	private static void save(PersistenceAction action, Collection<? extends Resource> resources, IProgressMonitor monitor) throws Exception {
		if (action != null && resources != null && !resources.isEmpty()) {
			ProgressMonitorUtil.beginTask(monitor, resources.size());
			for (Resource resource : resources) {
				if (resource != null && resource.getURI() != null) {
					if (action == PersistenceAction.DELETE) {
						URIConverter uriConverter = getURIConverter(resource);
						uriConverter.delete(resource.getURI(), null);
					} else {
						resource.save(null);
					}
				}

				ProgressMonitorUtil.checkIfCanceled(monitor);
				ProgressMonitorUtil.worked(monitor, 1);
			}
			ProgressMonitorUtil.done(monitor);
		}
	}

*/	/**
	 * Modifies the 'setting' defined by the specified eObject and feature. The given value is
	 * added if the feature is {@link EStructuralFeature#isMany() many} or is set if the feature is
	 * not many.
	 *
	 * @param eObject
	 * @param feature
	 * @param index the index for the object to be added or a negative number to
	 *        indicate that it should be appended to a many feature
	 *        (this parameter is not used if the feature is not many)
	 * @param value
	 * @return the index of the object, 0 if the feature is not many, or -1 if the setting was not
	 *         modified
	 */
/*	public static int setOrAddValue(EObject eObject, EStructuralFeature feature, int index, Object value) {
		if (eObject != null && feature != null && feature.isChangeable()) {
			if (feature.isMany()) {
				@SuppressWarnings({"rawtypes", "unchecked"})
				List<Object> list = (List)eObject.eGet(feature, true);
				if (index < 0) {
					list.add(value);
					index = list.size() - 1;
				} else {
					list.add(index, value);
				}
				return index;
			}

			eObject.eSet(feature, value);
			return 0;
		}
		return -1;
	}
*/
	/**
	 * Modifies the 'setting' defined by the specified eObject and feature. The given value is
	 * removed if the feature is {@link EStructuralFeature#isMany() many} or the feature is unset
	 * if it is not many.
	 *
	 * @param eObject
	 * @param feature
	 * @param index the index for the object to be remove or a negative number to
	 *        indicate that the last element of a many feature should be removed
	 *        (this parameter is not used if the feature is not many)
	 * @return the removed value or null
	 */
/*	public static Object unsetOrRemoveValue(EObject eObject, EStructuralFeature feature, int index) {
		if (eObject != null && feature != null && feature.isChangeable()) {
			Object value;
			if (feature.isMany()) {
				@SuppressWarnings({"rawtypes", "unchecked"})
				List<Object> list = (List)eObject.eGet(feature, true);

				if (index < 0) {
					value = list.remove(index);
				} else {
					index = list.size() - 1;
					value = list.remove(index);
				}
			} else {
				value = eObject.eGet(feature);
				eObject.eUnset(feature);
			}

			return value;
		}
		return null;
	}

	public static String getLastSegment(java.net.URI uri) {
		if (uri != null) {
			String path  = uri.getPath();
			int index = path.lastIndexOf('/');
			if (index >= 0) {
				StringBuilder sb = new StringBuilder();
				for (int i = index+1; i < path.length(); ) {
					int codePoint = path.codePointAt(i);
					if (Character.isAlphabetic(codePoint)) {
						sb.appendCodePoint(codePoint);
					}
					i += Character.charCount(codePoint);
				}

				if (sb.length() > 0) {
					return sb.toString();
				}
			}
		}
		return null;
	}
	*/
}
