//Copyright © 2015, Oracle and/or its affiliates. All rights reserved.

package util;

import java.text.CollationKey;
import java.text.Collator;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Random;
import java.util.function.Function;

/**
 * String utilities
 *
 * @author Marcelo Paternostro (marcelo.paternostro@oracle.com)
 */
public final class StringUtil {
	private StringUtil() {
		// Not instantiable
	}

	/**
	 * A empty string array
	 */
	public static final String[] EMPTY_STRING_ARRAY = new String[0];

	private static final char[] RANDON_STRING_CHARACTERS;
	static {
		// Lower 'a' to 'z' must be first set of characters
		StringBuilder sb = new StringBuilder(80);
		for (char c = 'a'; c <= 'z'; c++) {
			sb.append(c);
		}

		for (char c = 'A'; c <= 'Z'; c++) {
			sb.append(c);
		}

		for (char c = '0'; c <= '9'; c++) {
			sb.append(c);
		}

		RANDON_STRING_CHARACTERS = new char[sb.length()];
		sb.getChars(0, sb.length(), RANDON_STRING_CHARACTERS, 0);
	}

	/**
	 * @param string
	 * @return <code>true</code> if the string is empty or <code>null</code>
	 */
	public static boolean isEmpty(String string) {
		return string == null || string.length() == 0;
	}

	/**
	 * @param string
	 * @return <code>true</code> if the string is neither empty nor <code>null</code>
	 */
	public static boolean isNotEmpty(String string) {
		return string != null && string.length() > 0;
	}

	/**
	 * @param string
	 * @return a not-<code>null</code> string
	 *         (a <code>null</code> string is returned an empty string)
	 */
	public static String toNotNullString(String string) {
		return string == null ? "" : string; //$NON-NLS-1$
	}

	/**
	 * @param string
	 * @return <code>null</code> or a not-empty string
	 */
	public static String toNotEmptyString(String string) {
		return isEmpty(string) ? null : string;
	}

	/**
	 * @param string
	 * @return
	 */
	public static String toTrimmedString(String string) {
	    return isEmpty(string) ? string : string.trim();
	}

	/**
     * Checks that the specified string is not {@code null} and not
     * empty. This method is designed primarily for doing parameter validation in methods
     * and constructors, as demonstrated below:
     * <blockquote><pre>
     * public Foo(String bar) {
     *     this.bar = StringUtil.requireNonEmpty(bar);
     * }
     * </pre></blockquote>
	 * @param string
	 * @return the specified string if not empty
	 * @throws IllegalArgumentException if the string is empty
	 */
	public static String requireNonEmpty(String string) throws IllegalArgumentException {
		if (isEmpty(string)) {
			throw new IllegalArgumentException("string cannot be null nor empty"); //$NON-NLS-1$
		}
		return string;
	}

	/**
	 * @param strings
	 * @return a not-<code>null</code> list with the strings sorted according to the current locale's
	 *         {@link Collator collator}
	 * @see #collatorSort(Iterable, Collator)
	 */
	public static List<String> collatorSort(Iterable<String> strings) {
		return collatorSort(strings, (Collator)null);
	}

	/**
	 * @param strings
	 * @param locale a locale or <code>null</code> for the current locale
	 * @return a not-<code>null</code> list with the strings sorted according to the
	 *         specified locale's {@link Collator collator}
	 * @see #collatorSort(Iterable, Collator)
	 */
	public static List<String> collatorSort(Iterable<String> strings, Locale locale) {
		return collatorSort(strings, locale == null ? null : Collator.getInstance(locale));
	}

	/**
	 * Sorts the specified strings using a {@link Collator collator}.
	 *
	 * <p>This method is potentially faster than passing the strings as {@link List list} and the
	 * collator as the {@link java.util.Comparator comparator} to
	 * {@link java.util.Collections#sort(List, java.util.Comparator)}, especially for
	 * a 'not-small' amount of strings.</p>
	 *
	 * <p>Empty and <code>null</code> string are placed at the end of the returned list.</p>
	 *
	 * @param strings
	 * @param collator a collator or <code>null</code> for the current locale's collator
	 * @return a not-<code>null</code> list with the strings sorted according to the
	 *         specified collator
	 */
	public static List<String> collatorSort(Iterable<String> strings, Collator collator) {
		if (strings != null) {
			if (collator == null) {
				collator = Collator.getInstance();
			}

			List<CollationKey> collationKeys = new LinkedList<>();
			List<String> emptyStrings = new LinkedList<>();

			for (String string : strings) {
				if (isNotEmpty(string)) {
					CollationKey collationKey = collator.getCollationKey(string);
					collationKeys.add(collationKey);
				} else {
					emptyStrings.add(string);
				}
			}

			if (!collationKeys.isEmpty()) {
				Collections.sort(collationKeys);

				List<String> list = new ArrayList<>(collationKeys.size() + emptyStrings.size());
				for (CollationKey collationKey : collationKeys) {
					list.add(collationKey.getSourceString());
				}
				list.addAll(emptyStrings);

				return list;
			}

			if (!emptyStrings.isEmpty()) {
				return emptyStrings;
			}
		}
		return new ArrayList<>(0);
	}

	/**
	 *
	 * @param elements
	 * @param stringConverter
	 * @return a not-<code>null</code> list with the elements sorted according to how the
	 *         current locale's {@link Collator collator} sorts their string representation
	 * @see #collatorSort(Iterable, Function, Collator)
	 */
	public static <T> List<T> collatorSort(Iterable<? extends T> elements, Function<T, String> stringConverter) {
		return collatorSort(elements, stringConverter, (Collator)null);
	}

	/**
	 *
	 * @param elements
	 * @param stringConverter
	 * @param locale a locale or <code>null</code> for the current locale
	 * @return a not-<code>null</code> list with the elements sorted according to how the
	 *         specified locale's {@link Collator collator} sorts their string representation
	 * @see #collatorSort(Iterable, Function, Collator)
	 */
	public static <T> List<T> collatorSort(Iterable<? extends T> elements, Function<T, String> stringConverter, Locale locale) {
		return collatorSort(elements, stringConverter, locale == null ? null : Collator.getInstance(locale));
	}

	/**
	 * Uses a {@link Collator collator} to sorts the specified elements by their string representations
	 * computed as defined by the {@link #toString(Object, Function)} method.
	 *
	 * <p>Elements with empty and <code>null</code> string representations are placed at the end of the
	 * returned list.</p>
	 *
	 * @param elements
	 * @param stringConverter a converter to compute the element's string representation or
	 *                        <code>null</code> if the element's {@link Object#toString() toString}
	 *                        method is to be used
	 * @param collator a collator or <code>null</code> for the current locale's collator
	 * @return a not-<code>null</code> list with the elements sorted according to how the
	 *         {@link Collator collator} sorts their string representation
	 */
	public static <T> List<T> collatorSort(Iterable<? extends T> elements, Function<T, String> stringConverter, Collator collator) {
		if (elements != null) {
			List<T> emptyStringElements = new LinkedList<>();
			Map<String, List<T>> map = new HashMap<>();
			for (T element : elements) {
				String string = toString(element, stringConverter);
				if (isNotEmpty(string)) {
					List<T> list = map.get(string);
					if (list == null) {
						list = new ArrayList<>();
						map.put(string, list);
					}
					list.add(element);
				} else {
					emptyStringElements.add(element);
				}
			}

			if (!map.isEmpty()) {
				List<T> list = new ArrayList<>(map.size() + emptyStringElements.size());

				List<String> sorted = collatorSort(map.keySet(), collator);
				for (String string : sorted) {
					list.addAll(map.get(string));
				}
				list.addAll(emptyStringElements);

				return list;
			}

			if (!emptyStringElements.isEmpty()) {
				return emptyStringElements;
			}
		}

		return new ArrayList<>(0);
	}

	/**
	 * Returns a string representation for the specified object that is computed by
	 * the given <code>stringConverter</code>. If the converter is <code>null</code>, the
	 * method returns <code>null</code> if the object is <code>null</code> or the object's {@link Object#toString()}.
	 *
	 * @param object
	 * @param stringConverter <code>null</code> or a function to convert the object into a string.
	 * @return <code>null</code> or a string representing the object
	 */
	public static <T> String toString(T object, Function<? super T, String> stringConverter) {
		return
			stringConverter != null ? stringConverter.apply(object) :
			object != null ? object.toString() :
			null;
	}

	/**
	 * @param length
	 * @return <code>null</code> if length is negative or a random string of the specified
	 *         length in which all characters are either digits, or lower or upper case letters
	 *         (the string is guaranteed to start with a lower case letter between 'a' and 'z')
	 */
	public static String computeRandonString(int length) {
		if (length < 0) {
			return null;
		}

		if (length == 0) {
			return ""; //$NON-NLS-1$
		}

		Random random = new Random();
		char[] chars = new char[length];
		for (int i = 0; i < length; i++) {
			chars[i] = RANDON_STRING_CHARACTERS[random.nextInt(i == 0 ? 26 : RANDON_STRING_CHARACTERS.length)];
		}
		return new String(chars);
	}
}
